==> download.php?FileName=terralib-5.3.2.zip: src/terralib/edit/qt/tools/CreatePolygonTool.cpp <==
/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/edit/qt/tools/CreatePolygonTool.cpp

  \brief This class implements a concrete tool to create polygons.
*/

// TerraLib
#include "../../../common/STLUtils.h"
#include "../../../geometry.h"
#include "../../../dataaccess/dataset/ObjectId.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../qt/widgets/canvas/MapDisplay.h"
#include "../../../qt/widgets/Utils.h"
#include "../../RepositoryManager.h"
#include "../../Utils.h"
#include "../Renderer.h"
#include "../Utils.h"
#include "CreatePolygonTool.h"

// Qt
#include <QMessageBox>
#include <QMouseEvent>
#include <QPainter>
#include <QPixmap>

// STL
#include <cassert>
#include <memory>

te::edit::CreatePolygonTool::CreatePolygonTool(te::qt::widgets::MapDisplay* display, const te::map::AbstractLayerPtr& layer, const QCursor& cursor, 
                                               bool showVertexes, QObject* parent)
  : GeometriesUpdateTool(display, layer.get(), parent),
    m_continuousMode(false),
    m_isFinished(false),
    m_showVertexes(showVertexes),
    m_stack(UndoStackManager::getInstance())
{
  setCursor(cursor);

  // Signals & slots
  connect(m_display, SIGNAL(extentChanged()), SLOT(onExtentChanged()));

  draw();
}

te::edit::CreatePolygonTool::~CreatePolygonTool()
{
}

bool te::edit::CreatePolygonTool::mousePressEvent(QMouseEvent* e)
{
  if (e->button() != Qt::LeftButton)
    return false;

  if(m_isFinished) // Is Finished?! So, start again...
  {
    clear();
    m_isFinished = false;
  }

  if (m_coords.empty())
    emit geometriesStarted();

  QPointF pw = m_display->transform(GetPosition(e));

  te::gm::Coord2D coord = te::gm::Coord2D(pw.x(), pw.y());

  TrySnap(coord, m_display->getSRID());

  m_coords.push_back(coord);

  storeUndoCommand();

  return true;
}

bool te::edit::CreatePolygonTool::mouseMoveEvent(QMouseEvent* e)
{
  if(m_coords.size() < 1 || m_isFinished)
    return false;

  QPointF pos = GetPosition(e);
  pos += QPointF(0.0001, 0.0001); // To avoid collinear points on polygon

  QPointF pw = m_display->transform(pos);

  te::gm::Coord2D coord = te::gm::Coord2D(pw.x(), pw.y());

  TrySnap(coord, m_display->getSRID());

  m_lastPos = te::gm::Coord2D(coord.x, coord.y);

  if (e->buttons() & Qt::LeftButton)
  {
    double tol = 0.0;

    m_continuousMode = true;

    double scaleValue = m_display->getScale();
    double digitValue = m_display->getDigitalization();

    // Calculate of the digitalization (Code from Old Version: TA4)
    tol = (((scaleValue*digitValue) / 1000) / 1000);
    if (tol <= 0.0001347)
      tol = 0.0001347;

    if (GetDistance(m_coords.at(m_coords.size() - 1), coord) >= tol)
    {
      m_coords.push_back(coord);
      storeUndoCommand();
    }
  }
  else
  {
    m_coords.push_back(coord);

    m_continuousMode = false;
  }

  draw();

  return false;
}

bool te::edit::CreatePolygonTool::mouseDoubleClickEvent(QMouseEvent* e)
{
  try
  {
    if (e->button() != Qt::LeftButton)
      return false;

    return editingFinished();
  }
  catch (std::exception& e)
  {
    QMessageBox::critical(m_display, tr("Error"), QString(tr("Could not create.") + " %1.").arg(e.what()));
    return false;
  }
}

bool te::edit::CreatePolygonTool::mouseReleaseEvent(QMouseEvent* e)
{
  try
  {
    if (e->button() != Qt::RightButton)
      return false;

    return editingFinished();
  }
  catch (std::exception& e)
  {
    QMessageBox::critical(m_display, tr("Error"), QString(tr("Could not create.") + " %1.").arg(e.what()));
    return false;
  }
}

void te::edit::CreatePolygonTool::draw()
{
  const te::gm::Envelope& env = m_display->getExtent();
  if(!env.isValid())
    return;

  // Clear!
  QPixmap* draft = m_display->getDraftPixmap();
  draft->fill(Qt::transparent);

  // Initialize the renderer
  Renderer& renderer = Renderer::getInstance();
  renderer.begin(draft, env, m_display->getSRID());

  // Draw the layer edited geometries
  renderer.drawRepository(m_layer->getId(), env, m_display->getSRID());

  if(!m_coords.empty())
  {
    // Draw the geometry being created
    if(m_coords.size() < 3)
      drawLine();
    else if (m_stack.m_currentIndex > -1)
      drawPolygon();
    
    if(m_continuousMode == false)
      m_coords.pop_back();
  }

  renderer.end();

  m_display->repaint();
}

void te::edit::CreatePolygonTool::drawPolygon()
{
  // Build the geometry
  te::gm::Geometry* polygon = buildPolygon();

  // Draw the current geometry and the vertexes
  Renderer& renderer = Renderer::getInstance();
  renderer.draw(polygon, m_showVertexes);

  delete polygon;
}

void te::edit::CreatePolygonTool::drawLine()
{
  // Build the geometry
  te::gm::Geometry* line = buildLine();

  // Draw the current geometry and the vertexes
  Renderer& renderer = Renderer::getInstance();
  renderer.draw(line, m_showVertexes);

  delete line;
}

void te::edit::CreatePolygonTool::clear()
{
  delete m_feature;
  m_feature = 0;

  m_coords.clear();
}

te::gm::Geometry* te::edit::CreatePolygonTool::buildPolygon()
{
  // Build the geometry
  te::gm::LinearRing* ring = new te::gm::LinearRing(m_coords.size() + 1, te::gm::LineStringType);
  for(std::size_t i = 0; i < m_coords.size(); ++i)
    ring->setPoint(i, m_coords[i].x, m_coords[i].y);
  ring->setPoint(m_coords.size(), m_coords[0].x, m_coords[0].y); // Closing...

  te::gm::Polygon* polygon = new te::gm::Polygon(1, te::gm::PolygonType);
  polygon->setRingN(0, ring);
  polygon->setSRID(m_display->getSRID());

  return polygon;
}

te::gm::Geometry* te::edit::CreatePolygonTool::buildLine()
{
  te::gm::LineString* line = new te::gm::LineString(m_coords.size(), te::gm::LineStringType);
  for(std::size_t i = 0; i < m_coords.size(); ++i)
    line->setPoint(i, m_coords[i].x, m_coords[i].y);

  return line;
}

void te::edit::CreatePolygonTool::onExtentChanged()
{
  if(m_coords.empty())
    return;

  m_coords.push_back(m_lastPos);
}

void te::edit::CreatePolygonTool::storeUndoCommand()
{
  if (m_coords.size() < 3)
    return;

  if (m_feature == 0)
    m_feature = new Feature();

  m_feature->setGeometry(buildPolygon());
  m_feature->setOperationTypeId(INSERT);

  m_stack.addWatch(m_feature->clone());

  QUndoCommand* command = new AddCommand(m_display, m_layer, m_feature->clone()->getId());
  connect(dynamic_cast<AddCommand*>(command), SIGNAL(commandFeedback()), SLOT(onCommandFeedback()));

  m_stack.addUndoStack(command);
}

void te::edit::CreatePolygonTool::onCommandFeedback()
{
  m_isFinished = false;

  RepositoryManager& repo = RepositoryManager::getInstance();

  if (repo.hasIdentify(m_layer->getId(), m_feature->getId()) == true)
  {
    m_feature = repo.getFeature(m_layer->getId(), m_feature->getId())->clone();
    repo.removeFeature(m_layer->getId(), m_feature->getId());
  }

  m_coords.clear();

  GetCoordinates(m_feature->getGeometry(), m_coords);
  m_coords.pop_back();
}

void te::edit::CreatePolygonTool::resetVisualizationTool()
{
  clear();
}

bool te::edit::CreatePolygonTool::editingFinished()
{
  if (m_coords.size() < 3) // Can not stop yet...
    return false;

  m_isFinished = true;

  if (m_feature)
    storeFeature();

  emit geometriesEdited();

  return true;
}

void te::edit::CreatePolygonTool::storeFeature()
{
  if (m_feature == 0)
    return;

  RepositoryManager::getInstance().addFeature(m_layer->getId(), m_feature->clone());
}

==> download.php?FileName=terralib-5.4.0.zip: src/terralib/edit/qt/tools/CreatePolygonTool.cpp <==
/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/edit/qt/tools/CreatePolygonTool.cpp

  \brief This class implements a concrete tool to create polygons.
*/

// TerraLib
#include "../../../common/STLUtils.h"
#include "../../../geometry.h"
#include "../../../dataaccess/dataset/ObjectId.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../qt/widgets/canvas/MapDisplay.h"
#include "../../../qt/widgets/Utils.h"
#include "../../RepositoryManager.h"
#include "../../Utils.h"
#include "../Renderer.h"
#include "../Utils.h"
#include "CreatePolygonTool.h"

// Qt
#include <QMessageBox>
#include <QMouseEvent>
#include <QPainter>
#include <QPixmap>

// STL
#include <cassert>
#include <memory>

te::edit::CreatePolygonTool::CreatePolygonTool(te::qt::widgets::MapDisplay* display, const te::map::AbstractLayerPtr& layer, const QCursor& cursor, 
                                               bool showVertexes, QObject* parent)
  : GeometriesUpdateTool(display, layer.get(), parent),
    m_continuousMode(false),
    m_isFinished(false),
    m_showVertexes(showVertexes),
    m_stack(UndoStackManager::getInstance())
{
  setCursor(cursor);

  // Signals & slots
  connect(m_display, SIGNAL(extentChanged()), SLOT(onExtentChanged()));

  draw();
}

te::edit::CreatePolygonTool::~CreatePolygonTool() = default;

bool te::edit::CreatePolygonTool::mousePressEvent(QMouseEvent* e)
{
  if (e->button() != Qt::LeftButton)
    return false;

  if(m_isFinished) // Is Finished?! So, start again...
  {
    clear();
    m_isFinished = false;
  }

  if (m_coords.empty())
    emit geometriesStarted();

  QPointF pw = m_display->transform(GetPosition(e));

  te::gm::Coord2D coord = te::gm::Coord2D(pw.x(), pw.y());

  TrySnap(coord, m_display->getSRID());

  m_coords.push_back(coord);

  storeUndoCommand();

  return true;
}

bool te::edit::CreatePolygonTool::mouseMoveEvent(QMouseEvent* e)
{
  if(m_coords.size() < 1 || m_isFinished)
    return false;

  QPointF pos = GetPosition(e);
  pos += QPointF(0.0001, 0.0001); // To avoid collinear points on polygon

  QPointF pw = m_display->transform(pos);

  te::gm::Coord2D coord = te::gm::Coord2D(pw.x(), pw.y());

  TrySnap(coord, m_display->getSRID());

  m_lastPos = te::gm::Coord2D(coord.x, coord.y);

  if (e->buttons() & Qt::LeftButton)
  {
    double tol = 0.0;

    m_continuousMode = true;

    double scaleValue = m_display->getScale();
    double digitValue = m_display->getDigitalization();

    // Calculate of the digitalization (Code from Old Version: TA4)
    tol = (((scaleValue*digitValue) / 1000) / 1000);
    if (tol <= 0.0001347)
      tol = 0.0001347;

    if (GetDistance(m_coords.at(m_coords.size() - 1), coord) >= tol)
    {
      m_coords.push_back(coord);
      storeUndoCommand();
    }
  }
  else
  {
    m_coords.push_back(coord);

    m_continuousMode = false;
  }

  draw();

  return false;
}

bool te::edit::CreatePolygonTool::mouseDoubleClickEvent(QMouseEvent* e)
{
  try
  {
    if (e->button() != Qt::LeftButton)
      return false;

    return editingFinished();
  }
  catch (std::exception& e)
  {
    QMessageBox::critical(m_display, tr("Error"), QString(tr("Could not create.") + " %1.").arg(e.what()));
    return false;
  }
}

bool te::edit::CreatePolygonTool::mouseReleaseEvent(QMouseEvent* e)
{
  try
  {
    if (e->button() != Qt::RightButton)
      return false;

    return editingFinished();
  }
  catch (std::exception& e)
  {
    QMessageBox::critical(m_display, tr("Error"), QString(tr("Could not create.") + " %1.").arg(e.what()));
    return false;
  }
}

void te::edit::CreatePolygonTool::draw()
{
  const te::gm::Envelope& env = m_display->getExtent();
  if(!env.isValid())
    return;

  // Clear!
  QPixmap* draft = m_display->getDraftPixmap();
  draft->fill(Qt::transparent);

  // Initialize the renderer
  Renderer& renderer = Renderer::getInstance();
  renderer.begin(draft, env, m_display->getSRID());

  // Draw the layer edited geometries
  renderer.drawRepository(m_layer->getId(), env, m_display->getSRID());

  if(!m_coords.empty())
  {
    // Draw the geometry being created
    if(m_coords.size() < 3)
      drawLine();
    else if (m_stack.m_currentIndex > -1)
      drawPolygon();
    
    if(m_continuousMode == false)
      m_coords.pop_back();
  }

  renderer.end();

  m_display->repaint();
}

void te::edit::CreatePolygonTool::drawPolygon()
{
  // Build the geometry
  te::gm::Geometry* polygon = buildPolygon();

  // Draw the current geometry and the vertexes
  Renderer& renderer = Renderer::getInstance();
  renderer.draw(polygon, m_showVertexes);

  delete polygon;
}

void te::edit::CreatePolygonTool::drawLine()
{
  // Build the geometry
  te::gm::Geometry* line = buildLine();

  // Draw the current geometry and the vertexes
  Renderer& renderer = Renderer::getInstance();
  renderer.draw(line, m_showVertexes);

  delete line;
}

void te::edit::CreatePolygonTool::clear()
{
  delete m_feature;
  m_feature = nullptr;

  m_coords.clear();
}

te::gm::Geometry* te::edit::CreatePolygonTool::buildPolygon()
{
  // Build the geometry
  te::gm::LinearRing* ring = new te::gm::LinearRing(m_coords.size() + 1, te::gm::LineStringType);
  for(std::size_t i = 0; i < m_coords.size(); ++i)
    ring->setPoint(i, m_coords[i].x, m_coords[i].y);
  ring->setPoint(m_coords.size(), m_coords[0].x, m_coords[0].y); // Closing...

  te::gm::Polygon* polygon = new te::gm::Polygon(1, te::gm::PolygonType);
  polygon->setRingN(0, ring);
  polygon->setSRID(m_display->getSRID());

  return polygon;
}

te::gm::Geometry* te::edit::CreatePolygonTool::buildLine()
{
  te::gm::LineString* line = new te::gm::LineString(m_coords.size(), te::gm::LineStringType);
  for(std::size_t i = 0; i < m_coords.size(); ++i)
    line->setPoint(i, m_coords[i].x, m_coords[i].y);

  return line;
}

void te::edit::CreatePolygonTool::onExtentChanged()
{
  if(m_coords.empty())
    return;

  m_coords.push_back(m_lastPos);
}

void te::edit::CreatePolygonTool::storeUndoCommand()
{
  if (m_coords.size() < 3)
    return;

  if (m_feature == nullptr)
    m_feature = new Feature();

  m_feature->setGeometry(buildPolygon());
  m_feature->setOperationTypeId(INSERT);

  m_stack.addWatch(m_feature->clone());

  QUndoCommand* command = new AddCommand(m_display, m_layer, m_feature->clone()->getId());
  connect(dynamic_cast<AddCommand*>(command), SIGNAL(commandFeedback()), SLOT(onCommandFeedback()));

  m_stack.addUndoStack(command);
}

void te::edit::CreatePolygonTool::onCommandFeedback()
{
  m_isFinished = false;

  RepositoryManager& repo = RepositoryManager::getInstance();

  if (repo.hasIdentify(m_layer->getId(), m_feature->getId()) == true)
  {
    m_feature = repo.getFeature(m_layer->getId(), m_feature->getId())->clone();
    repo.removeFeature(m_layer->getId(), m_feature->getId());
  }

  m_coords.clear();

  GetCoordinates(m_feature->getGeometry(), m_coords);
  m_coords.pop_back();
}

void te::edit::CreatePolygonTool::resetVisualizationTool()
{
  clear();
}

bool te::edit::CreatePolygonTool::editingFinished()
{
  if (m_coords.size() < 3) // Can not stop yet...
    return false;

  m_isFinished = true;

  if (m_feature)
    storeFeature();

  emit geometriesEdited();

  return true;
}

void te::edit::CreatePolygonTool::storeFeature()
{
  if (m_feature == nullptr)
    return;

  RepositoryManager::getInstance().addFeature(m_layer->getId(), m_feature->clone());
}
